/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/
#include "sysinc.h"
#include "child.h"
#include "util.h"
#define F_GETSIG 11

extern struct child_glob_struct cglob;

void childDelWatch(int fd) {
  int i;
  for (i=0; i<cglob.watchcount; i++) 
    if (fd == cglob.watch[i])
      break;
  if (i<cglob.watchcount) {
    if (i<cglob.watchcount-1)
      memmove(&cglob.watch[i], &cglob.watch[i+1], (cglob.watchcount-i-1)*4);
  }
  cglob.watchcount--;
}

void child_main(char * ipcid) {
  struct rlimit rlim;
  int i, newfd, bufsize=0, done=0, res, fd;
  unsigned int snsize;
  char buf[65536];
  fdinfo * fdi;
  if (getrlimit(RLIMIT_NOFILE, &rlim) || !rlim.rlim_max)
    rlim.rlim_max = FOPEN_MAX;
  signal(SIGINT, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGHUP, SIG_IGN);
  for (i=0;i<=rlim.rlim_max;i++) {
    newfd = dup(i);
    if (newfd >= 0) {
      close(newfd);
      cglob.fdcount++;
      cglob.fdlist = (fdinfo *) realloc(cglob.fdlist, cglob.fdcount * sizeof(fdinfo));
      fdi = &cglob.fdlist[cglob.fdcount-1];
      memset(fdi, 0, sizeof(*fdi));
      fdi->fd = i;
      fdi->flags = fcntl(i, F_GETFL);
      fdi->owner = fcntl(i, F_GETOWN);
      fdi->iosig = fcntl(i, F_GETSIG);
      snsize = sizeof(fdi->sockname);
      res = getsockname(i, (struct sockaddr *) &fdi->sockname, &snsize);
      if (res)
	fdi->type = 0;
      else
	fdi->type = FDTYPE_SOCKET;
      snsize=sizeof(fdi->peername);
      getpeername(i, (struct sockaddr *) &fdi->peername, &snsize);
    }
  }
  /* Now see if we can connect to main */
  cglob.ipc = ipc_attach(ipcid);
  if (!cglob.ipc) {
    /* We couldn't. */
    exit(0);
  }
  do {
    bufsize = sizeof(buf)-8;
    if ((res = ipc_readblock(cglob.ipc, &buf, &bufsize)) > 0) {
      if (*(int *) &buf == CMD_EXIT) {
	done = 1;
      } else if (*(int *) &buf == CMD_FDLIST) {
	int res;
	char * tmp = malloc(sizeof(fdinfo) * cglob.fdcount + 4);
	*(int *) tmp = RESP_FDLIST;
	memcpy(tmp + 4, cglob.fdlist, cglob.fdcount * sizeof(fdinfo));
	res = ipc_sendblock(cglob.ipc, tmp, sizeof(fdinfo) * cglob.fdcount + 4);
	free(tmp);
      } else if (*(int *) buf == CMD_CLOSE) {
	close(*(int *) &buf[4]);
      } else if (*(int *) buf == CMD_SEND) {
	write(*(int *) &buf[4], (char *)&buf[8], bufsize-8);
      } else if (*(int *) buf == CMD_WATCH) {
	fd = *(int *) &buf[4];
	for (i=0; i<cglob.watchcount; i++) 
	  if (fd == cglob.watch[i])
	    break;
	if (i>=cglob.watchcount) {
	  cglob.watch[cglob.watchcount]=fd;
	  cglob.watchcount++;
	}
      } else if (*(int *) buf == CMD_STOPWATCH) {
	fd = *(int *) &buf[4];
	childDelWatch(fd);
      } else {
      }
    } else if (res<0) {
      exit(0);
    }
    if (cglob.watchcount) {
      int hfd = 0;
      struct timeval tv={0, 0};
      fd_set rfds, efds;
      FD_ZERO(&rfds);
      FD_ZERO(&efds);
      for (i=0;i<cglob.watchcount;i++) {
	if (hfd < cglob.watch[i])
	  hfd = cglob.watch[i];
	FD_SET(cglob.watch[i], &rfds);
	FD_SET(cglob.watch[i], &efds);
      }
      if (select(hfd+1, &rfds, 0, &efds, &tv)) {
	for (i=0;i<cglob.watchcount;i++) {
	  if (FD_ISSET(cglob.watch[i], &rfds)) {
	    char buf[4096];
	    int bytes = read(cglob.watch[i], &buf[8], sizeof(buf)-8);
	    if (bytes>0) {
	      *(int *) &buf = RESP_DATA;
	      *(int *) &buf[4] = cglob.watch[i];
	      ipc_sendblock(cglob.ipc, &buf, bytes+8);
	    } else {
	      *(int *) &buf = RESP_CLOSE;
	      *(int *) &buf[4] = cglob.watch[i];
	      ipc_sendblock(cglob.ipc, &buf, 8);
	      close(cglob.watch[i]);
	      childDelWatch(cglob.watch[i]);
	    }
	  }
	  if (FD_ISSET(cglob.watch[i], &efds)) {
	    *(int *) &buf = RESP_CLOSE;
	    *(int *) &buf[4] = cglob.watch[i];
	    ipc_sendblock(cglob.ipc, &buf, 8);
	    close(cglob.watch[i]);
	    childDelWatch(cglob.watch[i]);
	  }
	}
      }
    }
    usleep(100);
  } while (!done);
  ipc_close(cglob.ipc);
}
