/*

This file is (C) Copyright 2002 Erik S. Johansen <einride@einride.org>, all 
rights reserved.

If you work for or represent an organisation, agency or corporation that is 
wholly or partially funded by a government -OR- you are in any way associated 
with the company Cyber Operations, the following licence applies.

You may not use this file or any derivatives of it for any purpose whatsoever,
including reading beyond this comment, redistribution or compiling, without 
explicit written permission from the copyright holder. 

If you do not work for or represent an organisation, agency or corporation 
that is wholly or partially funded by a government -AND- you are not in any 
way associated with the company Cyber Operations, the GNU General Public 
License, as described in the accompanying file GPL, applies.

*/
#include "inject.h"
#include "trace.h"
#include "opcode.h"
#include "main.h"
#include "ui.h"
#include "debug.h"

extern struct app_arg_struct args;
extern struct app_glob_struct glob;
extern char * binname;

int inject_prepare() {
  if (!glob.process->flags) {
    trace_wait(glob.process, 0);
    trace_step(glob.process);
  }
  if (!TraceStopped(glob.process)) {
    ui_showmessage(UI_LEVEL_NORMAL, "Waiting up to 5 seconds for target to exit syscall");
    if (trace_wait(glob.process, 5)) {
      ui_showmessage(UI_LEVEL_ERROR, "Process didn't exit current syscall - trying to provoke a stop");
      kill(args.pid, SIGCHLD);
      if (trace_wait(glob.process, 15)) {
	ui_showmessage(UI_LEVEL_ERROR, "Failed stopping process (%i)", errno);
	trace_resume(glob.process, 0);
	return 1;
      }
    }
  }

  /* Save registers */
  if (trace_getregisters(glob.process, &glob.savedregs)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed reading registers (%i)", errno);
    trace_resume(glob.process, 0);
    return 1;
  }
  debug_dumpregisters(0);

  /* Save the code we'll overwrite */
  if (trace_read(glob.process, glob.insertionpoint, &glob.savedcode, sizeof(glob.savedcode))) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed reading code (%i)", errno);
    trace_resume(glob.process, 0);
    return 1;
  }
  return 0;
}

int inject_process(pcode code) {
  int bufsize, sig;
  char * buf;
  struct registers r;

  bufsize = code_size(code);
  bufsize = bufsize + 64 - (bufsize % 64);
  buf = (char *) malloc(bufsize); 
  memset(buf, 0x90, bufsize);

  code_write(code, (int) buf, buf);
  code_write(code, glob.insertionpoint, buf);
  code_kill(code);
  if (trace_write(glob.process, glob.insertionpoint, buf, bufsize)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed writing code (%i)", errno);
    trace_resume(glob.process, 0);
    return 1;
  }
  ui_showmessage(UI_LEVEL_VDEBUG, "Dumping original registers:");
  debug_dumpregisters(0);
  ui_showmessage(UI_LEVEL_DEBUG, "Changing EIP to 0x%08x", glob.insertionpoint + 4);
  memcpy(&r, &glob.savedregs, sizeof(r));
  r.eip = glob.insertionpoint + 4;
  if (trace_setregisters(glob.process, &r)) {
    ui_showmessage(UI_LEVEL_ERROR, "Error setting targets registers (%i)", errno);
    kill(args.pid, SIGSEGV);
    return 1;
  }

  do {
    if (trace_resume(glob.process, 0)) {
      ui_showmessage(UI_LEVEL_ERROR, "Failed resuming target (%i)", errno);
      kill(args.pid, SIGSEGV);
      return 1;
    }
    if (trace_wait(glob.process, 15)) {
      ui_showmessage(UI_LEVEL_ERROR, "Failed wait when executing injected code (%i)", errno);
      kill(args.pid, SIGSEGV);
      return 1;    
    }
    if (trace_getregisters(glob.process, &r)) {
      ui_showmessage(UI_LEVEL_ERROR, "Failed reading registers (%i)", errno);
      kill(args.pid, SIGSEGV);
      return 1;
    }
    sig = trace_readsignal(glob.process);
    if (sig==SIGSEGV) {
      ui_showmessage(UI_LEVEL_ERROR, "Target faulted");
      debug_dumpregisters(0);
      kill(args.pid, SIGSEGV);
      return 1;
    }
  } while (sig != SIGTRAP);
  ui_showmessage(UI_LEVEL_VDEBUG, "Dumping registers after execution of injected code:");
  debug_dumpregisters(0);
  return 0;
}

int inject_finish() {
  //  struct registers r;
  //  int done;
  if (trace_write(glob.process, glob.insertionpoint, &glob.savedcode, sizeof(glob.savedcode))) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed restoring code (%i)", errno);
    kill(args.pid, SIGSEGV);
    return 1;    
  }
  if (trace_setregisters(glob.process, &glob.savedregs)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed restoring registers (%i)", errno);
    kill(args.pid, SIGSEGV);
    return 1;    
  }
  if (trace_step(glob.process)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed singlestepping after restoring code (%i)", errno);
    return 1;   
  }
  if (trace_wait(glob.process, 15)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed waiting after restoring code (%i)", errno);
    return 1;   
  }
  return 0;
}

/*************************************************************************
 * Funcs making target do specific tasks by injecting code
 *************************************************************************/
int inject_forcechild(char * ipcid) {
  /* Insert code:
     if (!fork())
       execve(self, ("binname", "-child", ipcid), null)
     else
       int3
  */
  pcode code;
  if (inject_prepare())
    return 1;

  /* Create code to inject */
  code = code_create();
  code_addins(code, "start", "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  /* fork() */
  code_addins(code, 0, "mov eax, imm", 0x02);
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "nop");

  /* check fork() retval */
  code_addins(code, 0, "test eax, eax");
  code_addins(code, 0, "jnz", "done");
  code_addins(code, 0, "nop");

  /* we're child */
  /* nanosleep for 1 sec */
  code_addins(code, 0, "mov eax, imm", 0xA2);
  code_addins(code, 0, "mov ecx, imm", 0);
  code_addins(code, 0, "mov ebx, imm", 10000000);
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "nop");

  /* execve */
  code_addins(code, 0, "mov eax, imm", 0x0b);
  code_addins(code, 0, "mov ebx, abs", "binname");
  code_addins(code, 0, "mov ecx, abs", "arglist");
  code_addins(code, 0, "mov edx, imm", 0);
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "nop");

  /* exit (should never be called but better than a messed up fork()'ed target if something fails) */
  code_addins(code, 0, "mov eax, imm", 0x01);
  code_addins(code, 0, "mov ebx, imm", 0x00);
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "nop");

  /* done */  
  code_addins(code, "done", "nop");
  code_addins(code, 0, "int3");

  code_addabsref(code, "arglist", "binname");
  code_addabsref(code, 0, "arg2");
  code_addabsref(code, 0, "arg3");
  code_adddword(code, 0, 0);
  
  code_addraw(code, "binname", binname, strlen(binname)+1);
  code_addraw(code, "arg2", "-child", 7);
  code_addraw(code, "arg3", ipcid, strlen(ipcid)+1);
  code_adddword(code, 0, 0);

  if (inject_process(code)) {
    inject_finish();
    return 1;
  }

  return inject_finish();
}

int inject_forceconnect(char * sockname, int * sock) {
  /* Insert code:
     RedirectedFD = socket(AF_UNIX, SOCK_STREAM, 0);
     sa.sun_family = AF_UNIX;
     sa.sun_path = sockname;
     connect(RedirectedFD, &sa, sizeof(sa));
  */
  pcode code;
  int res;
  struct registers r;
  struct sockaddr_un sa;

  if (inject_prepare())
    return 1;

  memset(&sa, 0, sizeof(sa));
  sa.sun_family = AF_UNIX;
  strcpy(sa.sun_path, sockname);

  /* Create code to inject */
  code = code_create();
  code_addins(code, "start", "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "mov edx, imm", 0xFFFFFFFF);
  /* socket() */
  code_addins(code, 0, "push imm4", 0);
  code_addins(code, 0, "push imm4", SOCK_STREAM);
  code_addins(code, 0, "push imm4", AF_UNIX);
  code_addins(code, 0, "mov ecx, esp");
  code_addins(code, 0, "mov eax, imm", 0x66);
  code_addins(code, 0, "mov ebx, imm", 0x01);
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "nop");
  
  /* check socket() retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_socket");
  code_addins(code, 0, "nop");

  /* Save the socket fd*/
  code_addins(code, 0, "mov ebp, eax");

  /* connect() */
  code_addins(code, 0, "push imm4", sizeof(sa));
  code_addins(code, 0, "push abs", "sockaddr");
  code_addins(code, 0, "push eax");
  code_addins(code, 0, "mov eax, imm", 0x66);
  code_addins(code, 0, "mov ebx, imm", 0x03);
  code_addins(code, 0, "mov ecx, esp");
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "nop");
  
  /* check the connect retval */
  code_addins(code, 0, "test eax, eax");
  code_addins(code, 0, "jnz", "failed_connect");

  /* edx=0 -> all ok */
  code_addins(code, 0, "mov edx, imm", 0);
  code_addins(code, 0, "int3");

  /* edx=1 -> socket() failed */
  code_addins(code, "failed_socket", "mov edx, imm", 1);
  code_addins(code, 0, "int3");

  /* edx=2 -> connect() failed */
  code_addins(code, "failed_connect", "mov edx, imm", 2);
  code_addins(code, 0, "int3");

  /* The sockname_un struct */
  code_addraw(code, "sockaddr", &sa, sizeof(sa));

  if (inject_process(code)) {
    inject_finish();
    return 1;
  }
  if (trace_getregisters(glob.process, &r)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed reading registers (%i)", errno);
    return 1;
  }
  res = r.edx;
  if (!res)
    *sock = r.ebp;

  if (inject_finish())
    return 1;

  return res;
}

int inject_forcefcntlsetfl(int fd, int flags) {
  /* Insert code:
     fcntl(fd, F_SETFL, flags);
  */
  pcode code;

  if (inject_prepare())
    return 1;

  /* Create code to inject */
  code = code_create();
  code_addins(code, "start", "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "mov edx, imm", flags);
  code_addins(code, 0, "mov ebx, imm", fd);
  code_addins(code, 0, "mov ecx, imm", F_SETFL);
  code_addins(code, 0, "mov eax, imm", 0x37);
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "nop");
  code_addins(code, 0, "int3");
  
  if (inject_process(code)) {
    inject_finish();
    return 1;
  }
  return inject_finish();
}

int inject_forceclose(int fd) {
  /* Insert code:
     close(fd);
  */
  pcode code;

  if (inject_prepare())
    return 1;
  
  /* Create code to inject */
  code = code_create();
  code_addins(code, "start", "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");

  /* close(fd) */
  code_addins(code, 0, "mov ebx, imm", fd);
  code_addins(code, 0, "mov eax, imm", 0x6);
  code_addins(code, 0, "int", 0x80);
  code_addins(code, 0, "int3");

  if (inject_process(code)) {
    inject_finish();
    return 1;
  }

  return inject_finish();
}

int inject_forceredirectfd(int fd, int sock, int * savedfd) {
  /* Insert code:
     dup(fd);
     mov ebp, eax
     close(fd)
     dup2(sock, fd);
     close(sock);

     *savedfd = targets EBP
  */
  pcode code;
  struct registers r;
  int res;

  if (inject_prepare())
    return 1;
  
  /* Create code to inject */
  code = code_create();
  code_addins(code, "start", "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");

  /* dup(fd) */
  code_addins(code, 0, "mov ebx, imm", fd);
  code_addins(code, 0, "mov eax, imm", 0x29);
  code_addins(code, 0, "int", 0x80);
  /* check dup retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_dup");

  /* save new FD */
  code_addins(code, 0, "mov ebp, eax");

  /* close(fd) */
  code_addins(code, 0, "mov ebx, imm", fd);
  code_addins(code, 0, "mov eax, imm", 0x6);
  code_addins(code, 0, "int", 0x80);
  /* check retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_close_1");

  /* dup2(sock, fd) */
  code_addins(code, 0, "mov ecx, imm", fd);
  code_addins(code, 0, "mov ebx, imm", sock);
  code_addins(code, 0, "mov eax, imm", 0x29);
  code_addins(code, 0, "int", 0x80);
  /* check dup2 retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_dup2");

  /* close(sock) */
  code_addins(code, 0, "mov ebx, imm", sock);
  code_addins(code, 0, "mov eax, imm", 0x6);
  code_addins(code, 0, "int", 0x80);
  /* check retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_close_2");

  /* ok, EDX=0 */
  code_addins(code, 0, "mov edx, imm", 0);
  code_addins(code, 0, "int3");

  /* dup failed, EDX = 1 */
  code_addins(code, "failed_dup", "mov edx, imm", 1);
  code_addins(code, 0, "int3");

  /* close(fd) failed, EDX = 2 */
  code_addins(code, "failed_close_1", "mov edx, imm", 2);
  code_addins(code, 0, "int3");

  /* dup2 failed, EDX = 3 */
  code_addins(code, "failed_dup2", "mov edx, imm", 3);
  code_addins(code, 0, "int3");

  /* close(sock) failed, EDX = 4 */
  code_addins(code, "failed_close_2", "mov edx, imm", 4);
  code_addins(code, 0, "int3");

  if (inject_process(code)) {
    inject_finish();
    return 1;
  }

  if (trace_getregisters(glob.process, &r)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed reading registers (%i)", errno);
    return 1;
  }
  res = r.edx;
  *savedfd = r.ebp;
  if (inject_finish())
    return 1;
  return res;
}

int inject_forcerestorefd(int savedfd, int fd) {
  /* Insert code:
     close(fd);
     dup2(savedfd, fd);
     close(savedfd);
  */
  pcode code;
  struct registers r;
  int res;

  if (inject_prepare())
    return 1;
  
  /* Create code to inject */
  code = code_create();
  code_addins(code, "start", "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");
  code_addins(code, 0, "nop");

  /* close(fd) */
  code_addins(code, 0, "mov ebx, imm", fd);
  code_addins(code, 0, "mov eax, imm", 0x6);
  code_addins(code, 0, "int", 0x80);
  /* check retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_close_1");
  
  /* dup2(savedfd, fd) */
  code_addins(code, 0, "mov ecx, imm", fd);
  code_addins(code, 0, "mov ebx, imm", savedfd);
  code_addins(code, 0, "mov eax, imm", 0x29);
  code_addins(code, 0, "int", 0x80);
  /* check dup2 retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_dup2");

  /* close(savedfd) */
  code_addins(code, 0, "mov ebx, imm", savedfd);
  code_addins(code, 0, "mov eax, imm", 0x6);
  code_addins(code, 0, "int", 0x80);
  /* check retval */
  code_addins(code, 0, "cmp eax, imm", 0);
  code_addins(code, 0, "jl", "failed_close_2");

  /* ok, EDX=0 */
  code_addins(code, 0, "mov edx, imm", 0);
  code_addins(code, 0, "int3");

  /* close(fd) failed, EDX = 1 */
  code_addins(code, "failed_close_1", "mov edx, imm", 1);
  code_addins(code, 0, "int3");

  /* dup2 failed, EDX = 2 */
  code_addins(code, "failed_dup2", "mov edx, imm", 2);
  code_addins(code, 0, "int3");

  /* close(sock) failed, EDX = 3 */
  code_addins(code, "failed_close_2", "mov edx, imm", 3);
  code_addins(code, 0, "int3");

  if (inject_process(code)) {
    inject_finish();
    return 1;
  }

  if (trace_getregisters(glob.process, &r)) {
    ui_showmessage(UI_LEVEL_ERROR, "Failed reading registers (%i)", errno);
    return 1;
  }
  res = r.edx;
  if (inject_finish())
    return 1;
  return res;
}

